@@ -1,6 +1,6 @@
 # -*- test-case-name: foolscap.test.test_negotiate -*-
 
-from twisted.python import log
+import time
 from twisted.python.failure import Failure
 from twisted.internet import protocol, reactor
 
@@ -8,7 +8,10 @@ from foolscap import broker, referenceable, vocab
 from foolscap.eventual import eventually
 from foolscap.tokens import SIZE_LIMIT, ERROR, \
      BananaError, NegotiationError, RemoteNegotiationError
+from foolscap.ipb import DeadReferenceError
 from foolscap.banana import int2b128
+from foolscap.logging import log
+from foolscap.logging.log import NOISY, OPERATIONAL, WEIRD, UNUSUAL, CURIOUS
 
 crypto_available = False
 try:
@@ -160,7 +163,9 @@ class Negotiation(protocol.Protocol):
     SERVER_TIMEOUT = 60 # you have 60 seconds to complete negotiation, or else
     negotiationTimer = None
 
-    def __init__(self):
+    def __init__(self, logparent=None):
+        self._logparent = log.msg("Negotiation started", parent=logparent,
+                                  facility="foolscap.negotiation")
         for i in range(self.minVersion, self.maxVersion+1):
             assert hasattr(self, "evaluateNegotiationVersion%d" % i), i
             assert hasattr(self, "acceptDecisionVersion%d" % i), i
@@ -200,8 +205,21 @@ class Negotiation(protocol.Protocol):
         # connectionLost() message is finally delivered.
         self.failureReason = None
 
+    def log(self, *args, **kwargs):
+        # we log as NOISY by default, because nobody should hear about
+        # negotiation unless it goes wrong.
+        if 'parent' not in kwargs:
+            kwargs['parent'] = self._logparent
+        if 'facility' not in kwargs:
+            kwargs['facility'] = "foolscap.negotiation"
+        if 'level' not in kwargs:
+            kwargs['level'] = log.NOISY
+        return log.msg(*args, **kwargs)
+
     def initClient(self, connector, targetHost):
         # clients do connectTCP and speak first with a GET
+        self.log("initClient: to target %s" % connector.target,
+                 target=connector.target.getTubID())
         self.isClient = True
         self.tub = connector.tub
         self.brokerClass = self.tub.brokerClass
@@ -212,9 +230,15 @@ class Negotiation(protocol.Protocol):
         self.wantEncryption = bool(self.target.encrypted
                                    or self.tub.myCertificate)
         self.options = self.tub.options.copy()
+        tubID = self.target.getTubID()
+        # note that UnauthenticatedTubs will never have a record here
+        slave_record = self.tub.slave_table.get(tubID, ("none",0))
+        assert isinstance(slave_record, tuple), slave_record
+        self.negotiationOffer['last-connection'] = "%s %s" % slave_record
 
     def initServer(self, listener):
         # servers do listenTCP and respond to the GET
+        self.log("initServer", listener=repr(listener))
         self.isClient = False
         self.listener = listener
         self.options = self.listener.options.copy()
@@ -240,7 +264,7 @@ class Negotiation(protocol.Protocol):
     def debug_doTimer(self, name, timeout, call, *args):
         if (self.options.has_key("debug_slow_%s" % name) and
             not self.debugTimers.has_key(name)):
-            log.msg("debug_doTimer(%s)" % name)
+            self.log("debug_doTimer(%s)" % name)
             t = reactor.callLater(timeout, self.debug_fireTimer, name)
             self.debugTimers[name] = (t, [(call, args)])
             cb = self.options["debug_slow_%s" % name]
@@ -306,18 +330,14 @@ class Negotiation(protocol.Protocol):
         # Hello block.
         req = []
         if self.target.encrypted:
-            if self.debugNegotiation:
-                log.msg("sendPlaintextClient: GET for tubID %s"
-                        % self.target.tubID)
+            self.log("sendPlaintextClient: GET for tubID %s" %
+                     self.target.tubID)
             req.append("GET /id/%s HTTP/1.1" % self.target.tubID)
         else:
-            if self.debugNegotiation:
-                log.msg("sendPlaintextClient: GET for no tubID")
+            self.log("sendPlaintextClient: GET for no tubID")
             req.append("GET /id/ HTTP/1.1")
         req.append("Host: %s" % self.targetHost)
-        if self.debugNegotiation:
-            log.msg("sendPlaintextClient: wantEncryption=%s" %
-                    self.wantEncryption)
+        self.log("sendPlaintextClient: wantEncryption=%s" % self.wantEncryption)
         if self.wantEncryption:
             req.append("Upgrade: TLS/1.0")
         else:
@@ -355,9 +375,9 @@ class Negotiation(protocol.Protocol):
             del self.negotiationTimer
 
     def dataReceived(self, chunk):
-        if self.debugNegotiation:
-            log.msg("dataReceived(isClient=%s,phase=%s,options=%s): '%s'"
-                    % (self.isClient, self.receive_phase, self.options, chunk))
+        self.log("dataReceived(isClient=%s,phase=%s,options=%s): %r"
+                 % (self.isClient, self.receive_phase, self.options, chunk),
+                 level=NOISY)
         if self.receive_phase == ABANDONED:
             return
 
@@ -394,8 +414,7 @@ class Negotiation(protocol.Protocol):
 
         except Exception, e:
             why = Failure()
-            if self.debugNegotiation:
-                log.msg("negotiation had exception: %s" % why)
+            self.log("negotiation had exception", failure=why)
             if isinstance(e, RemoteNegotiationError):
                 pass # they've already hung up
             else:
@@ -404,8 +423,8 @@ class Negotiation(protocol.Protocol):
                 if isinstance(e, NegotiationError):
                     errmsg = str(e)
                 else:
-                    log.msg("negotiation had internal error:")
-                    log.msg(why)
+                    self.log("negotiation had internal error:", failure=why,
+                             level=UNUSUAL)
                     errmsg = "internal server error, see logs"
                 errmsg = errmsg.replace("\n", " ").replace("\r", " ")
                 if self.send_phase == PLAINTEXT:
@@ -461,24 +480,25 @@ class Negotiation(protocol.Protocol):
             # probably a web browser
             raise BananaError("not right")
         targetTubID = url[4:]
-        if self.debugNegotiation:
-            log.msg("handlePLAINTEXTServer: targetTubID='%s'" % targetTubID)
+        self.log("handlePLAINTEXTServer: targetTubID='%s'" % targetTubID,
+                 level=NOISY)
         if targetTubID == "":
             targetTubID = None
         if isSubstring("Upgrade: TLS/1.0\r\n", header):
             wantEncrypted = True
         else:
             wantEncrypted = False
-        if self.debugNegotiation:
-            log.msg("handlePLAINTEXTServer: wantEncrypted=%s" % wantEncrypted)
+        self.log("handlePLAINTEXTServer: wantEncrypted=%s" % wantEncrypted,
+                 level=NOISY)
         # we ignore the rest of the lines
 
         if wantEncrypted and not crypto_available:
             # this is a confused client, or a bad URL: if we don't have
             # crypto, we couldn't have created a pb:// URL.
-            log.msg("Negotiate.handlePLAINTEXTServer: client wants "
-                    "encryption for TubID=%s but we have no crypto, "
-                    "hanging up on them" % targetTubID)
+            self.log("Negotiate.handlePLAINTEXTServer: client wants "
+                     "encryption for TubID=%s but we have no crypto, "
+                     "hanging up on them" % targetTubID,
+                     level=UNUSUAL)
             # we could just not offer the encryption, but they won't be happy
             # with the results, since they wanted to connect to a specific
             # TubID.
@@ -540,8 +560,7 @@ class Negotiation(protocol.Protocol):
         raise NotImplementedError # TODO
 
     def handlePLAINTEXTClient(self, header):
-        if self.debugNegotiation:
-            log.msg("handlePLAINTEXTClient: header='%s'" % header)
+        self.log("handlePLAINTEXTClient: header='%s'" % header)
         lines = header.split("\r\n")
         tokens = lines[0].split()
         # TODO: accept a 303 redirect
@@ -564,9 +583,8 @@ class Negotiation(protocol.Protocol):
         # which might actually involve a TLS-encrypted session if that's what
         # the client wanted, but if it isn't then we just "upgrade" to
         # nothing and change modes.
-        if self.debugNegotiation:
-            log.msg("startENCRYPTED(isClient=%s, encrypted=%s)" %
-                    (self.isClient, encrypted))
+        self.log("startENCRYPTED(isClient=%s, encrypted=%s)" %
+                 (self.isClient, encrypted))
         if encrypted:
             self.startTLS(self.tub.myCertificate)
         self.encrypted = encrypted
@@ -590,9 +608,12 @@ class Negotiation(protocol.Protocol):
             # identity.
             hello['my-tub-id'] = self.myTubID
 
-        if self.debugNegotiation:
-            log.msg("Negotiate.sendHello (isClient=%s): %s" %
-                    (self.isClient, hello))
+        if self.tub:
+            IR = self.tub.getIncarnationString()
+            hello['my-incarnation'] = IR
+
+        self.log("Negotiate.sendHello (isClient=%s): %s" %
+                 (self.isClient, hello))
         self.sendBlock(hello)
 
 
@@ -635,9 +656,8 @@ class Negotiation(protocol.Protocol):
             - We are not the master: DECISION is None
         """
 
-        if self.debugNegotiation:
-            log.msg("evaluateHello(isClient=%s): offer=%s" % (self.isClient,
-                                                              offer,))
+        self.log("evaluateHello(isClient=%s): offer=%s" %
+                 (self.isClient, offer))
         if not offer.has_key('banana-negotiation-range'):
             if offer.has_key('banana-negotiation-version'):
                 msg = ("Peer is speaking foolscap-0.0.5 or earlier, "
@@ -733,8 +753,7 @@ class Negotiation(protocol.Protocol):
             # this is the most common case
             iAmTheMaster = myTubID > theirTubID
 
-        if self.debugNegotiation:
-            log.msg("iAmTheMaster: %s" % iAmTheMaster)
+        self.log(format="iAmTheMaster: %(master)s", master=iAmTheMaster)
 
         decision, params = None, None
 
@@ -743,17 +762,55 @@ class Negotiation(protocol.Protocol):
             # a decision block.
             self.send_phase = DECIDING
             decision = {}
+            params = {}
             # combine their 'offer' and our own self.negotiationOffer to come
             # up with a 'decision' to be sent back to the other end, and the
             # 'params' to be used on our connection
 
-            # first, do we continue with this connection? we might
-            # have an existing connection for this particular tub
+            # first, do we continue with this connection? we might have an
+            # existing connection for this particular tub
+
             if theirTubRef and theirTubRef in self.tub.brokers:
-                # there is an existing connection, so drop this one
-                if self.debugNegotiation:
-                    log.msg(" abandoning the connection: we already have one")
-                raise NegotiationError("Duplicate connection")
+                # there is an existing connection.. we might want to prefer
+                # this new offer, because the old connection might be stale
+                # (NAT boxes and laptops that disconnect abruptly are two
+                # ways for a single process to disappear silently and then
+                # reappear with a different IP address).
+                lp = self.log("got offer for an existing connection",
+                              level=UNUSUAL)
+                existing = self.tub.brokers[theirTubRef]
+                acceptOffer = self.compareOfferAndExisting(offer, existing, lp)
+                if acceptOffer:
+                    # drop the old one
+                    self.log("accepting new offer, dropping existing connection",
+                             parent=lp)
+                    err = DeadReferenceError("[%s] replaced by a new connection"
+                                             % theirTubRef.getShortTubID())
+                    why = Failure(err)
+                    existing.shutdown(why)
+                else:
+                    # reject the new one
+                    self.log("rejecting the offer: we already have one",
+                             parent=lp)
+                    raise NegotiationError("Duplicate connection")
+
+            if theirTubRef:
+                # generate a new seqnum, one higher than the last one we've
+                # used. Note that UnauthenticatedTubs all share the same
+                # index, so we leak certain information about how many
+                # connections we've established.
+                old_seqnum = self.tub.master_table.get(theirTubRef.getTubID(),
+                                                       0)
+                new_seqnum = old_seqnum + 1
+                new_slave_IR = offer.get('my-incarnation', None)
+                self.tub.master_table[theirTubRef.getTubID()] = new_seqnum
+                my_IR = self.tub.getIncarnationString()
+                decision['current-connection'] = "%s %s" % (my_IR, new_seqnum)
+                # these params will be copied into the Broker where we can
+                # retrieve them later, when we need to compare it against a new
+                # offer.
+                params['current-slave-IR'] = new_slave_IR
+                params['current-seqnum'] = new_seqnum
 
             # what initial vocab set should we use?
             theirVocabRange_s = offer.get("initial-vocab-table-range", "0 0")
@@ -771,9 +828,8 @@ class Negotiation(protocol.Protocol):
             decision['banana-decision-version'] = str(self.decision_version)
 
             # v1: handle vocab table index
-            params = { 'banana-decision-version': self.decision_version,
-                       'initial-vocab-table-index': vocab_index,
-                       }
+            params['banana-decision-version'] = self.decision_version
+            params['initial-vocab-table-index'] = vocab_index
 
         else:
             # otherwise, the other side gets to decide. The next thing they
@@ -783,8 +839,8 @@ class Negotiation(protocol.Protocol):
 
         if iAmTheMaster:
             # I am the master, so I send the decision
-            if self.debugNegotiation:
-                log.msg("Negotiation.sendDecision: %s" % decision)
+            self.log("Negotiation.sendDecision: %s" % decision,
+                     level=OPERATIONAL)
             # now we send the decision and switch to Banana. they might hang
             # up.
             self.sendDecision(decision, params)
@@ -805,6 +861,152 @@ class Negotiation(protocol.Protocol):
         # changes were made to the offer or decision blocks.
         return self.evaluateNegotiationVersion1(offer)
 
+    def compareOfferAndExisting(self, offer, existing, lp):
+        """Compare the new offer against the existing connection, and
+        decide which to keep.
+
+        @return: True to accept the new offer, False to stick with the
+                 existing connection.
+        """
+
+        def log(*args, **kwargs):
+            if 'parent' not in kwargs:
+                kwargs['parent'] = lp
+            return self.log(*args, **kwargs)
+
+        existing_slave_IR = existing.current_slave_IR
+        existing_seqnum = existing.current_seqnum
+
+        log(format="existing connection has slave_IR=%(slave_IR)s, seqnum=%(seqnum)s",
+            slave_IR=existing_slave_IR, seqnum=existing_seqnum)
+
+        # TESTING: force handle-old stuff
+        #lp2 = log("TESTING: forcing use of handle-old logic")
+        #return self.handle_old(offer, existing, 60, lp2)
+
+        # step one: does the inbound offer have a my-incarnation header? If
+        # not, this is an older peer (<foolscap-0.1.7). We use
+        # offer.get("my-incarnation") instead of "my-incarnation" in offer
+        # so that unit tests can cause a client to send an empty string to
+        # simulate the earlier version.
+        if not offer.get("my-incarnation") or "last-connection" not in offer:
+            # TODO: new servers send my-incarnation but not last-connection
+
+            # this is an old peer (foolscap 0.1.7 or earlier), which won't
+            # give us enough information to make some of the decisions below.
+            # We reject the offer to avoid connection flap, and the
+            # situtation won't be worse than it was in 0.1.7 .
+            lp2 = log("pre-0.2.0 peer detected (no my-incarnation"
+                      " or last-connection", level=CURIOUS)
+            if self.tub._handle_old_duplicate_connections is not False:
+                # but if we've been configured to do better (with the
+                # 60-second age heuristic), do that.
+                self.log("using handle-old-duplicate-connections", parent=lp2)
+                threshold = self.tub._handle_old_duplicate_connections
+                return self.handle_old(offer, existing, threshold, lp2)
+            return False # reject the offer
+
+        if offer["my-incarnation"] != existing_slave_IR:
+            # this offer is from a different invocation of the peer than we
+            # think we're currently talking to. That means the slave has
+            # restarted since we made our connection, so clearly our
+            # connection is stale. Accept the offer.
+            log("offer is from different peer incarnation than existing")
+            return True # accept
+
+        pieces = offer['last-connection'].split()
+        offer_master_IR = pieces[0]
+        offer_master_seqnum = int(pieces[1])
+
+        if offer_master_IR == "none":
+            # the peer doesn't remember talking to anybody: they don't think
+            # they've ever been connected to us. We disagree, and we remember
+            # their incarnation record. So they must have made an initial
+            # attempt to connect to us (their first), we accepted their
+            # connection, and the decision message got lost or hasn't arrived
+            # yet. The most likely situation is that this is one of the
+            # parallel connections (one per hint), for which we want to
+            # reject their offer. The less likely situation is that they
+            # heard our initial connection setup but the decision message got
+            # lost, in which case we entry the "no reconnects until TCP gives
+            # up" state.
+            log("peer doesn't remember talking to us")
+            return False # reject
+
+        if offer_master_IR != self.tub.getIncarnationString():
+            # the peer doesn't remember talking to us specifically, but they
+            # remember talking to one of our past lives. That means our last
+            # decision message didn't make it to them, and the last
+            # connection they *did* hear about was from one of our previous
+            # runs. Therefore our existing connection isn't viable, and we
+            # should accept their offer.
+            #
+            log("peer remembers talking to our past life")
+            return True # accept
+
+        # at this point, the offer's IR matches our own, so the seqnum is
+        # worth comparing
+        if offer_master_seqnum == existing_seqnum:
+            # the offer demonstrates that the client knows about the same
+            # connection that we do, and they made a new connection anyways.
+            # From this we can conclude that our connection is stale, so we
+            # should accept the offer.
+            log("peer knows about existing seqnum")
+            return True
+
+        if offer_master_seqnum < existing_seqnum:
+            # Possible ways to get here, most likely first
+            #  1: simultaneous parallel connections (multiple hints),
+            #     from a client who used to have an established connection
+            #     with us (so they're sending the right offer_master_IR).
+            #     Reject the offer to avoid connection flap.
+            #  2: client connected, but our decision got lost, they're still
+            #     living in the past. Reject the offer, we'll enter the
+            #     no-reconnect-until-TCP-gives-up state
+            #  3: crazy stalled message case, again we wait for TCP to expire
+
+            # more details on #2: the client connects successfully, then the
+            # client thinks the connection has been lost (but the server
+            # thinks it's still good), so the client reconnects, and this
+            # connection gets as far as the master making a decision, but the
+            # decision message is lost before it gets to the client. Then the
+            # client connects a third time, and now we're considering the
+            # third offer: the IRs are all the same, the attempt_id is
+            # different than our existing (2nd) connection, but the seqnum is
+            # older. In this case, we want to accept the new offer.
+
+
+            # more details on #3 (more rare): the client connects and loses
+            # the connection (as before), then the client connects a second
+            # time and gets as far as sending the offer when they time out,
+            # cancelling the negotiation already in progress (sending a FIN
+            # after the offer message) and triggering a third connection. The
+            # third connection somehow races ahead and completes negotiation
+            # before the 2nd-connection offer+FIN make it to the server. Now,
+            # finally, the offer arrives: we're now evaluating an
+            # out-of-order offer on a socket that's about to be closed.
+            # Ideally we'd like to reject this offer.
+
+            log("peer knows about old seqnum")
+            return False # reject
+
+        # offer_master_seqnum > existing_seqnum indicates something really
+        # weird has taken place.
+        log(format="offer_master_seqnum %(offer)d > existing_seqnum %(existing)d",
+            offer=offer_master_seqnum, existing=existing_seqnum, level=WEIRD)
+        return False # reject weirdness
+
+    def handle_old(self, offer, existing, threshold, lp):
+        # determine the age of the existing broker
+        age = time.time() - existing.creation_timestamp
+        if age < threshold:
+            self.log("the existing broker is too new (%d<%d), rejecting offer"
+                     % (age, threshold),
+                     parent=lp)
+            return False # reject the offer
+        self.log("the existing broker is old enough to replace", parent=lp)
+        return True # accept the offer
+
     def sendDecision(self, decision, params):
         if self.debug_doTimer("sendDecision", 1,
                               self.sendDecision, decision, params):
@@ -818,9 +1020,8 @@ class Negotiation(protocol.Protocol):
 
     def handleDECIDING(self, header):
         # this gets called on the non-master side
-        if self.debugNegotiation:
-            log.msg("handleDECIDING(isClient=%s): %s" % (self.isClient,
-                                                         header))
+        self.log("handleDECIDING(isClient=%s): %s" % (self.isClient, header),
+                 level=NOISY)
         if self.debug_doTimer("handleDECIDING", 1,
                               self.handleDECIDING, header):
             # for testing purposes, wait a moment before accepting the
@@ -838,8 +1039,7 @@ class Negotiation(protocol.Protocol):
         the negotiation from the server. The client must accept this decision
         (and return the connection parameters dict), or raise
         NegotiationError to hang up.negotiationResults."""
-        if self.debugNegotiation:
-            log.msg("Banana.acceptDecision: got %s" % decision)
+        self.log("Banana.acceptDecision: got %s" % decision, level=OPERATIONAL)
 
         version = decision.get('banana-decision-version')
         if not version:
@@ -872,6 +1072,25 @@ class Negotiation(protocol.Protocol):
             msg = ("Our hash for vocab-table-index %d (%s) does not match "
                    "your hash (%s)" % (vocab_index, our_hash, vocab_hash))
             raise NegotiationError(msg)
+
+        if self.theirTubRef in self.tub.brokers:
+            # we're the slave, so we need to drop our existing connection and
+            # use the one picked by the master
+            self.log("master told us to use a new connection, "
+                     "so we must drop the existing one", level=UNUSUAL)
+            err = DeadReferenceError("replaced by a new connection")
+            why = Failure(err)
+            self.tub.brokers[self.theirTubRef].shutdown(why)
+
+        current_connection = decision.get('current-connection')
+        if current_connection:
+            tubID = self.theirTubRef.getTubID()
+            if tubID != "<unauth>":
+                self.tub.slave_table[tubID] = tuple(current_connection.split())
+        else:
+            self.log("no current-connection in decision from %s" %
+                     self.theirTubRef, level=UNUSUAL)
+
         params = { 'banana-decision-version': ver,
                    'initial-vocab-table-index': vocab_index,
                    }
@@ -910,8 +1129,7 @@ class Negotiation(protocol.Protocol):
         # We use the MyOptions class to fix up the verify stuff: we request a
         # certificate from the client, but do not verify it against a list of
         # root CAs
-        if self.debugNegotiation:
-            log.msg("startTLS, client=%s" % self.isClient)
+        self.log("startTLS, client=%s" % self.isClient)
         kwargs = {}
         if cert:
             kwargs['privateKey'] = cert.privateKey.original
@@ -924,13 +1142,18 @@ class Negotiation(protocol.Protocol):
         # switch over to the new protocol (a Broker instance). This
         # Negotiation protocol goes away after this point.
 
-        if self.debugNegotiation:
-            log.msg("Negotiate.switchToBanana(isClient=%s)" % self.isClient)
-            log.msg(" params: %s" % (params,))
+        lp = self.log("Negotiate.switchToBanana(isClient=%s)" % self.isClient,
+                      level=NOISY)
+        self.log("params: %s" % (params,), parent=lp)
 
         self.stopNegotiationTimer()
 
-        b = self.brokerClass(params,
+        if self.isClient:
+            theirTubRef = self.target
+        else:
+            theirTubRef = self.theirTubRef
+
+        b = self.brokerClass(theirTubRef, params,
                              self.tub.keepaliveTimeout,
                              self.tub.disconnectTimeout,
                              )
@@ -949,19 +1172,16 @@ class Negotiation(protocol.Protocol):
 
         # finally let our Tub know that they can start using the new Broker.
         # This will wake up anyone who initiated an outbound connection.
-        if self.isClient:
-            theirTubRef = self.target
-        else:
-            theirTubRef = self.theirTubRef
         self.tub.brokerAttached(theirTubRef, b, self.isClient)
 
     def negotiationFailed(self):
         reason = self.failureReason
-        if self.debugNegotiation:
-            # TODO: consider logging this unconditionally.. it shouldn't
-            # happen very often, but if it does, it may take a long time to
-            # track down
-            log.msg("Negotiation.negotiationFailed: %s" % reason)
+        # TODO: consider logging this unconditionally.. it shouldn't happen
+        # very often, but if it does, it may take a long time to track down.
+        # ACTUALLY, parallel connection-hints cause the slower connection to
+        # hit here with a duplicate connection reason all the time.
+        self.log("Negotiation.negotiationFailed", failure=reason,
+                 level=OPERATIONAL)
         self.stopNegotiationTimer()
         if self.receive_phase != ABANDONED and self.isClient:
             eventually(self.connector.negotiationFailed, self.factory, reason)
@@ -969,7 +1189,7 @@ class Negotiation(protocol.Protocol):
         cb = self.options.get("debug_negotiationFailed_cb")
         if cb:
             # note that this gets called with a NegotiationError, not a
-            # Failure
+            # Failure. ACTUALLY: not true, gets a Failure
             eventually(cb, reason)
 
 # TODO: make sure code that examines self.receive_phase handles ABANDONED
@@ -978,9 +1198,17 @@ class TubConnectorClientFactory(protocol.ClientFactory, object):
     # this is for internal use only. Application code should use
     # Tub.getReference(url)
 
-    def __init__(self, tc, host):
+    noisy = False
+
+    def __init__(self, tc, host, logparent):
         self.tc = tc # the TubConnector
         self.host = host
+        self._logparent = logparent
+
+    def log(self, *args, **kwargs):
+        kwargs['parent'] = self._logparent
+        kwargs['facility'] = "foolscap.negotiation"
+        return log.msg(*args, **kwargs)
 
     def __repr__(self):
         # make it clear which remote Tub we're trying to connect to
@@ -1003,11 +1231,20 @@ class TubConnectorClientFactory(protocol.ClientFactory, object):
     def startedConnecting(self, connector):
         self.connector = connector
 
+    def startFactory(self):
+        self.log("Starting factory %r" % self)
+        return protocol.ClientFactory.startFactory(self)
+    def stopFactory(self):
+        self.log("Stopping factory %r" % self)
+        return protocol.ClientFactory.stopFactory(self)
+
     def disconnect(self):
+        self.log("told to disconnect")
         self.connector.disconnect()
 
     def buildProtocol(self, addr):
-        proto = self.tc.tub.negotiationClass() # this is usually Negotiation
+        nc = self.tc.tub.negotiationClass # this is usually Negotiation
+        proto = nc(self._logparent)
         proto.initClient(self.tc, self.host)
         proto.factory = self
         return proto
@@ -1016,7 +1253,7 @@ class TubConnectorClientFactory(protocol.ClientFactory, object):
         self.tc.clientConnectionFailed(self, reason)
 
 
-class TubConnector:
+class TubConnector(object):
     """I am used to make an outbound connection. I am given a target TubID
     and a list of locationHints, and I try all of them until I establish a
     Broker connected to the target. I will consider redirections returned
@@ -1032,9 +1269,6 @@ class TubConnector:
     progress are cancelled, and established connections (the ones which did
     *not* complete negotiation before the winning connection) have called
     their connectionLost() methods.
-
-    @param locationHints: the list of 'host:port' hints where the remote tub
-                          can be found.
     """
 
     failureReason = None
@@ -1042,9 +1276,22 @@ class TubConnector:
     timer = None
 
     def __init__(self, parent, tubref):
+        self._logparent = log.msg(format="TubConnector created from "
+                                  "%(fromtubid)s to %(totubid)s",
+                                  fromtubid=parent.tubID,
+                                  totubid=tubref.getTubID(),
+                                  level=OPERATIONAL,
+                                  facility="foolscap.connection")
         self.tub = parent
         self.target = tubref
-        self.remainingLocations = self.target.getLocations()
+        hints = []
+        # filter out the hints that we can actually use.. there may be
+        # extensions from the future sitting in this list
+        for h in self.target.getLocations():
+            if h[0] == "ipv4":
+                (host, port) = h[1:]
+                hints.append( (host, port) )
+        self.remainingLocations = hints
         # attemptedLocations keeps track of where we've already tried to
         # connect, so we don't try them twice.
         self.attemptedLocations = []
@@ -1056,6 +1303,17 @@ class TubConnector:
         # or because someone told us to give up).
         self.pendingConnections = {}
 
+    def __repr__(self):
+        s = object.__repr__(self)
+        s = s[:-1]
+        s += " from %s to %s>" % (self.tub.tubID, self.target.getTubID())
+        return s
+
+    def log(self, *args, **kwargs):
+        kwargs['parent'] = self._logparent
+        kwargs['facility'] = "foolscap.connection"
+        return log.msg(*args, **kwargs)
+
     def connect(self):
         """Begin the connection process. This should only be called once.
         This will either result in the successful Negotiation object invoking
@@ -1090,9 +1348,9 @@ class TubConnector:
             if location in self.attemptedLocations:
                 continue
             self.attemptedLocations.append(location)
-            host, port = location.split(":")
-            port = int(port)
-            f = TubConnectorClientFactory(self, host)
+            host, port = location
+            lp = self.log("connectTCP to %s" % (location,))
+            f = TubConnectorClientFactory(self, host, lp)
             c = reactor.connectTCP(host, port, f)
             self.pendingConnections[f] = c
             # the tcp.Connector that we get back from reactor.connectTCP will
@@ -1134,6 +1392,7 @@ class TubConnector:
         # this is called if protocol negotiation cannot be established, or if
         # the connection is closed for any reason prior to switching to the
         # Banana protocol
+        self.log("negotiationFailed for factory %s" % factory, failure=reason)
         assert isinstance(reason, Failure), \
                "Hey, %s isn't a Failure" % (reason,)
         if (not self.failureReason or
@@ -1148,6 +1407,7 @@ class TubConnector:
     def negotiationComplete(self, factory):
         # 'factory' has just completed negotiation, so abandon all the other
         # connection attempts
+        self.log("negotiationComplete, %s won" % factory)
         self.active = False
         if self.timer:
             self.timer.cancel()
@@ -1180,10 +1440,10 @@ class TubConnector:
         if (self.failureReason.check(RemoteNegotiationError) and
             isSubstring(self.failureReason.value.args[0],
                         "Duplicate connection")):
-            log.msg("TubConnector.checkForFailure: connection attempt "
-                    "failed because the other end decided ours was a "
-                    "duplicate connection, so we won't signal the "
-                    "failure here")
+            self.log("TubConnector.checkForFailure: connection attempt "
+                     "failed because the other end decided ours was a "
+                     "duplicate connection, so we won't signal the "
+                     "failure here")
             return
         self.failed()
 
@@ -1199,4 +1459,5 @@ class TubConnector:
             return
         # we have no more outstanding connections (either in progress or in
         # negotiation), so this connector is finished.
+        self.log("connectorFinished (%s)" % self)
         self.tub.connectorFinished(self)
